/*
 * syncml plugin - A syncml plugin for OpenSync
 * Copyright (C) 2005  Armin Bauer <armin.bauer@opensync.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */
 
#ifndef _SYNCML_COMMON_H
#define _SYNCML_COMMON_H
//#include <config.h>

#include <opensync/opensync.h>

#include <opensync/opensync-format.h>
#include <opensync/opensync-plugin.h>
#include <opensync/opensync-merger.h>
#include <opensync/opensync-context.h>
#include <opensync/opensync-data.h>
#include <opensync/opensync-helper.h>
#include <opensync/opensync-group.h>
#include <opensync/opensync-version.h>

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>

#include <glib.h>

#include <libsyncml/syncml.h>

#include <libsyncml/sml_auth.h>
#include <libsyncml/sml_devinf_obj.h>
#include <libsyncml/sml_ds_server.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

/* Define limits for large object support:
 *
 * MaxMsgSize => 10kB
 * MaxObjSize => 10MB
 */
#define OSYNC_PLUGIN_SYNCML_MAX_MSG_SIZE 10240
#define OSYNC_PLUGIN_SYNCML_MAX_OBJ_SIZE 10240000

#define SYNCML_PLUGIN_CONFIG_SANVERSION "SANVersion"
#define SYNCML_PLUGIN_CONFIG_WBXML "WBXML"
#define SYNCML_PLUGIN_CONFIG_ATCOMMAND "ATCommand"
#define SYNCML_PLUGIN_CONFIG_ATMANUFACTURER "ATManufacturer"
#define SYNCML_PLUGIN_CONFIG_ATMODEL "ATModel"
#define SYNCML_PLUGIN_CONFIG_IDENTIFIER "Identifier"
#define SYNCML_PLUGIN_CONFIG_USESTRINGTABLE "UseStringTable"
#define SYNCML_PLUGIN_CONFIG_USETIMEANCHOR "UseTimeAnchor"
#define SYNCML_PLUGIN_CONFIG_ONLYREPLACE "OnlyReplace"
#define SYNCML_PLUGIN_CONFIG_MAXMSGSIZE "MaxMsgSize"
#define SYNCML_PLUGIN_CONFIG_MAXOBJSIZE "MaxObjSize"
#define SYNCML_PLUGIN_CONFIG_ONLYLOCALTIME "OnlyLocaltime"

typedef enum {
	OSYNC_PLUGIN_SYNCML_COMMAND_UNKNOWN,
	OSYNC_PLUGIN_SYNCML_COMMAND_SEND_ALERT,
	OSYNC_PLUGIN_SYNCML_COMMAND_SEND_SYNC,
	OSYNC_PLUGIN_SYNCML_COMMAND_RECV_SYNC,
} OSyncPluginSyncmlDatastoreCommand;

typedef struct SmlPluginEnv {
	const char *bluetoothAddress;
	char *bluetoothChannel;
	const char *atCommand;
	const char *atManufacturer;
	const char *atModel;
	const char *identifier;
	SmlNotificationVersion version;
	osync_bool useWbxml;
	const char *username;
	const char *password;
	SmlBool useStringtable;
	SmlBool onlyReplace;
	SmlBool onlyLocaltime;
	SmlTransportConnectionType type;
	char *port;
	char *url;
	char *proxy;
	char *cafile;
	SmlBool useTimestampAnchor;
	
	unsigned int maxMsgSize;
	unsigned int maxObjSize;

	SmlBool gotDisconnect;
	SmlBool tryDisconnect;
	SmlBool doReconnect;
	
	OSyncPluginInfo *pluginInfo;
	char *anchor_path;
	char *devinf_path;

	GSource *source;
	GSourceFuncs *source_functions;

	GMainContext *context;
	
	SmlTransport *tsp;
	SmlAuthenticator *auth;
	SmlDevInf *devinf;
	SmlDevInf *remote_devinf;
	SmlDevInfAgent *agent;
	SmlManager *manager;
	SmlSession *session;
	
	SmlNotification *san;

	GList *databases;
	GList *ignoredDatabases;
	char *sessionUser;

	int num;

	SmlAuthType authType;
	osync_bool fakeDevice;
	SmlProtocolVersion syncmlVersion;
        char *fakeManufacturer;
        char *fakeModel;
        char *fakeSoftwareVersion;

	osync_bool isConnected;
	OSyncContext *connectCtx;
	OSyncContext *disconnectCtx;
	GMutex *connectMutex;

	/* This function pointer is necessary to start the second OMA DS session
	 * if the synchronization uses an OMA DS client.
	 */
	OSyncSinkConnectFn connectFunction;

	GMutex *managerMutex;
} SmlPluginEnv;

typedef struct SmlDatabase {
	SmlPluginEnv *env;
	SmlDsSession *session;
	SmlDsServer *server;
	OSyncObjFormat *objformat;
	const char *objformat_name;
	OSyncObjTypeSink *sink;
	const char *objtype;	
	const char *url;
	char *remoteNext;
	char *localNext;

	SmlDsSessionAlertCb dsSessionCallback;

	OSyncChange **syncChanges;
	OSyncContext **syncContexts;
	unsigned int pendingChanges;
	unsigned int pendingCommits;

	OSyncPluginSyncmlDatastoreCommand command;

	OSyncContext *syncModeCtx;
	OSyncContext *getChangesCtx;
	OSyncContext *commitCtx;
} SmlDatabase;

struct commitContext {
	OSyncContext *context;
	OSyncChange *change;
	SmlDatabase *database;
};

gboolean _sessions_prepare(GSource *source, gint *timeout_);

gboolean _sessions_check(GSource *source);

gboolean _sessions_dispatch(
			GSource *source, 
			GSourceFunc callback, 
			gpointer user_data);

void sync_done(void *data, OSyncPluginInfo *info, OSyncContext *ctx);

void disconnect(void *data, OSyncPluginInfo *info, OSyncContext *ctx);

SmlBool send_sync_message(
                        SmlDatabase *database,
                        void *func_ptr,
                        OSyncError **oserror);

SmlDatabase *syncml_config_parse_database(
			SmlPluginEnv *env,
			OSyncPluginResource *res,
			OSyncError **error);

SmlBool flush_session_for_all_databases(
			SmlPluginEnv *env,
			SmlBool activeDatabase,
			SmlError **error);

SmlDatabase *get_database_from_plugin_info(OSyncPluginInfo *info);

/* this is a helper function which adds an object to a GList */
/* the function guarantees that an object exists only once in */
/* this GList. No double entrees. */
GList *g_list_add(GList *databases, void *database);

void initEnv(SmlPluginEnv *env);

void finalize(void *data);

void safe_cfree(char **address);
void safe_free(gpointer *address);

void report_success_on_context(OSyncContext **ctx);
void report_error_on_context(OSyncContext **ctx, OSyncError **error, osync_bool cleanupError);

// FIXME: This is only a fast fix for SuSE.
// FIXME: Perhaps the functions should be renamed.
OSyncChangeType _to_osync_changetype(SmlChangeType type);
void set_session_user(SmlPluginEnv *env, const char* user);
unsigned int get_num_changes(OSyncChange **changes);

char *get_next_anchor(SmlDatabase *database, const char *last);

#endif //_SYNCML_COMMON_H
